#!/usr/bin/env python

import os,sys
import logging
import imp
import math
import collections
import numpy

tmp_path = "%s/git/versuchung/src"% os.environ["HOME"]
if os.path.exists(tmp_path):
    sys.path.append(tmp_path)


from versuchung.experiment import Experiment
from versuchung.types import String, Bool,List
from versuchung.database import Database
from versuchung.files import File, Directory, Executable
from versuchung.archives import GitArchive
from versuchung.execute import shell
from versuchung.tex import DatarefDict
from threading import Thread


class FailTrace(Experiment):
    inputs = {
        "coredos": GitArchive("/proj/i4gerrit/review_site/git/coredos.git"),
        "mpu": Bool(True),
    }
    outputs = {
        "traces": Directory("traces"),
        "stats": File("statistics.py"),

    }

    def configure_and_build(self, config, directory):
        extra = ""
        name = "base"
        if self.mpu.value:
            extra += " --mpu=yes"
            name  += "+mpu"
        else:
            extra += " --mpu=no"

        if config["encode_system"]:
            extra += " --encoded=yes"
            name += "+enc"
        else:
            extra += " --encoded=no"

        if config["specialize_system"]:
            extra += " --specialize=yes"
            name  += "+opt"
        else:
            extra += " --specialize=no"


        logging.info("Configuring...")
        shell("cd %s; ../new_build_env.py -a i386 -c "+ extra, directory)

        SGFLAGS = ["-fsse", "-fgcfg"]
        if config["insert_state_asserts"]:
            SGFLAGS.append("-fgen-asserts")
            name += "+ass"

        config["name"] = name

        logging.info("Building..." + str(config))
        logging.info(" ".join(SGFLAGS))
        shell("export SGFLAGS=%s; cd %s; make -j $(nproc) fail-trace-bench-coptermock",
              " ".join(SGFLAGS), directory)

    def run(self):
        logging.info("Cloning coredos...")

        with self.coredos as coredos_path:
            self.stats.copy_contents("generator/statistics.py")

            # Generate different variant to compile and to trace
            variants = []
            for flags in [(False, False, False),
                          (False, True, False),
                          (False, False, True),
                          (True, False, False),
                          (True, True, False),
                          (True, False, True)]:
                variants += [{"encode_system": flags[0],
                              "specialize_system": flags[1],
                              "insert_state_asserts": flags[2],
                          }]


            def VAR(variant):
                directory = "build" + str(id(variant))
                os.mkdir(directory)

                self.configure_and_build(variant, directory)

                shell("mv %s/fail-bench-coptermock %s/%s",
                      directory,
                      self.traces.path,
                      variant["name"])

                shell("cp %s/app/benchmark/coptermock/fail-bench-coptermock/stats.dict.py %s/%s",
                      directory,
                      self.traces.path,
                      variant["name"])


            threads = []
            for variant in variants:
                t = Thread(target = VAR, args=(variant,))
                t.start()
                threads.append(t)
            for t in threads:
                t.join()


class FailTraceRuntime(Experiment):
    inputs = {
        "gather": FailTrace("FailTrace"),
        "thesis-unencoded": FailTrace("from-thesis/GatherStatsDict-ab90a9c06290c39bd5a37db4584b298e"),
        "thesis-encoded": FailTrace("from-thesis/GatherStatsDict-d8c6cb931748d82e702c6f7618828de1"),
    }
    outputs = {"dref" : DatarefDict() }

    def fmtn(self,name):
        return name.replace("bcc1_", "").replace("_", "-")

    def key(self, keys, value, basename = "runtime"):
        self.dref["/".join([basename] + keys)] = value

    def work_on_benchmark(self, name, stats):
        def S(keys, value, **kwargs):
            self.key([self.fmtn(name)] + keys, value, **kwargs)

        combined_cycles = []
        cycles = collections.defaultdict(lambda : [])

        syscalls = 0
        syscall_opts = []

        codesize = []

        for x in stats.find_all('AtomicBasicBlock').values():
            if not "activations" in x:
                continue

            codesize.append(x["generated-codesize"])

            # Count the optimizations
            opts = [opt for opt in x.keys()
                    if opt.startswith("opt:")]
            syscalls += 1
            syscall_opts.append(len(opts))

            # If name is of the form: ABB[0-9]/lalala it is a syscalls
            assert '/' in x["_name"]
            syscall = x["_name"].split("/")[1]

            # We ignore the kickoff "system call" here because: The
            # compiler might put some code there that does not belong
            # to the system.  Also does kickoff is not necessary for
            # the OSEK semantic. The trace-analyze.py script does also
            # not always detect an empyt kickoff() sequence.
            if syscall == "kickoff":
                continue

            for act in x["activations"]:
                cycles[syscall].append(act["cycles"])
                combined_cycles.append(act["cycles"])

        # Here we output the number of system calls, and the number of
        # system calls that were optimized.
        S(["syscalls", "count"], syscalls, basename = "optimization")
        S(["syscalls", "optimized"], len(filter(None, syscall_opts)), basename = "optimization")


        for syscall, activations in cycles.items():
            S([syscall, "sum"],   sum(activations))
            S([syscall, "mean"], numpy.mean(activations))
            S([syscall, "count"], len(activations))
            S([syscall, "std"], numpy.std(activations))

        activations = combined_cycles
        S(["sum"],   sum(activations))
        S(["mean"],  numpy.mean(activations))
        S(["count"], len(activations))
        S(["std"],   numpy.std(activations))

        print name, len(combined_cycles), numpy.mean(combined_cycles)

        row = codesize
        S(["sum"],   sum(row), basename="codesize")
        S(["mean"],  numpy.mean(row), basename="codesize")
        S(["std"],   numpy.std(row), basename="codesize")

    def run(self):
        # Load the statistics module, that was imported from coredos
        # to read the statstics
        statistics = imp.load_source("statistics", self.gather.stats.path)

        self.matrix = collections.defaultdict(list)
        for tmp in self.gather.traces:
            name = tmp.basename
            stats = statistics.Statistics.load(tmp.path + "/stats.dict.py")
            self.work_on_benchmark(name, stats)

        # We use the partial scheduler data, I used in my master's
        # thesis. Since the revision there support for partial
        # schedulers was removed. Since it was a problem to compile
        # REALLY large applications.
        stats = statistics.Statistics.load(self.thesis_unencoded.path + "/dicts/fail-bench-coptermock")
        self.work_on_benchmark("base+mpu+opt+partial scheduler", stats)
        stats = statistics.Statistics.load(self.thesis_encoded.path + "/dicts/fail-bench-coptermock")
        self.work_on_benchmark("base+mpu+enc+opt+partial scheduler", stats)


class FailTraceImport(Experiment):
    inputs = {
        "trace": FailTrace("FailTrace"),
        "fail-tool-dir": Directory("/proj/i4danceos/tools/fail"),
    }

    def run(self):
        for trace in self.trace.traces:
            variant = "dosek/" + trace.basename
            if "opt" in variant:
                continue

            for (label, importer, importer_args) in [\
                                ("mem",    "MemoryImporter", []),
                                ("regs",   "RegisterImporter", []),
                                ("ip",     "RegisterImporter", ["--no-gp", "--ip"]),
                                ("flags",  "RegisterImporter", ["--no-gp", "--flags"]),
                                            ]:
                benchmark = label
                logging.info("Importing coredos/%s", )
                cmdline = "%(path)s/import-trace -v %(variant)s -b %(benchmark)s -i %(importer)s "\
                          + "-t %(trace)s -e %(elf)s %(args)s"
                shell(cmdline %\
                      {"path": self.fail_tool_dir.path,
                       "variant": variant,
                       "benchmark": benchmark,
                       "importer": importer,
                       "trace":  os.path.join(trace.path, "trace.pb"),
                       "elf":  os.path.join(trace.path, "fail-bench-coptermock"),
                       "args": " ".join(importer_args)})
            shell("%s/prune-trace -v %s -b %% -p basic",
                  self.fail_tool_dir.path, variant)

if __name__ == "__main__":
    import shutil, sys
    if len(sys.argv) > 1:
        action = sys.argv[1]; del sys.argv[1]
    actions = {"trace": FailTrace,
               "runtime": FailTraceRuntime,
               "import" : FailTraceImport,
               }

    if action in actions:
        experiment = actions[action]()
        dirname = experiment(sys.argv)
    else:
        print "No action to be taken, use: %s" % ", ".join(actions.keys())

